import importlib
import inspect
import json
import os
import sys
from json import JSONDecodeError
from types import ModuleType
from typing import Any, Type

from autopack.errors import AutoPackLoadError, AutoPackNotFoundError, AutoPackNotInstalledError
from autopack.pack import Pack
from autopack.pack_response import PackResponse


def find_or_create_autopack_dir(depth=0) -> str:
    """Try to find a suitable .autopack directory. Tries in this order:
    1. Directory specified in environment variable AUTOPACK_DIR
    2. Existing .autopack directory in current directory
    3. Existing .autopack directory up to 3 directories up
    4. Creates an .autopack directory in current directory
    """
    env_dir = os.environ.get("AUTOPACK_DIR")
    if env_dir:
        return env_dir

    autopack_dir = os.path.abspath(os.path.join(os.getcwd(), *[os.pardir] * depth, ".autopack"))

    if not os.path.exists(autopack_dir) or not os.path.isdir(autopack_dir):
        if depth > 3:
            os.makedirs(".autopack", exist_ok=True)
            return ".autopack"
        return find_or_create_autopack_dir(depth=depth + 1)

    return autopack_dir


def load_metadata_file() -> dict[str, Any]:
    """Return the parsed contents of the metadata file, returning an empty dict if not found or otherwise failed"""
    metadata_dir = find_or_create_autopack_dir()
    metadata_file = os.path.join(metadata_dir, "pack_metadata.json")

    if not os.path.exists(metadata_file):
        return {}

    with open(metadata_file, "r") as f:
        try:
            return json.load(f)
        except JSONDecodeError:
            return {}


def write_metadata_file(data: dict[str, Any]):
    metadata_dir = find_or_create_autopack_dir()
    metadata_file = os.path.join(metadata_dir, "pack_metadata.json")

    with open(metadata_file, "w+") as f:
        json.dump(data, f)


def find_module(pack_data: PackResponse) -> ModuleType:
    autopack_dir = find_or_create_autopack_dir()
    module_path = pack_data.module_path
    pack_module_path = os.path.join(autopack_dir, pack_data.pack_path())

    sys.path.insert(0, autopack_dir)
    sys.path.insert(0, pack_module_path)

    try:
        return importlib.import_module(module_path)
    finally:
        sys.path.remove(autopack_dir)
        sys.path.remove(pack_module_path)


def fetch_pack_object(pack_data: PackResponse, quiet=False) -> Pack:
    try:
        module = find_module(pack_data)
        if not module:
            message = (
                f"Pack {pack_data.pack_id} could not be found. Either it is misconfigured or .autopack directory not "
                f"found"
            )
            raise AutoPackNotFoundError(message)

        for _, obj in inspect.getmembers(module):
            if is_valid_pack(obj, pack_data.name):
                return Pack(**{"tool_class": obj, **pack_data.__dict__})

        message = f"Pack {pack_data.pack_id} found, but {pack_data.name} is not found in its module"
        raise AutoPackNotFoundError(message)
    except ModuleNotFoundError:
        message = (
            f"Pack {pack_data.pack_id} is available but not installed. To install: autopack install {pack_data.pack_id}"
        )
        if not quiet:
            print(message)
        raise AutoPackNotInstalledError(message)
    except (ImportError, AttributeError) as e:
        message = f"Error loading {pack_data.pack_id}: {e}"
        if not quiet:
            print(message)
        raise AutoPackLoadError(message)


def is_valid_pack(klass: Type, name: str):
    if not inspect.isclass(klass):
        return False

    if hasattr(klass, "__fields__"):
        name_field = klass.__fields__.get("name")
        if name_field and name_field.default:
            snaked_name = name_field.default.lower().replace(" ", "_")
            return name_field and snaked_name == name

    return False



import json

from langchain.chat_models.base import BaseChatModel
from langchain.schema import SystemMessage, BaseMessage

from autopack.api import pack_search

TOOL_SELECTION_PROMPT_TEMPLATE = """Given the task and the tool list provided below, please return a JSON object identifying which tools would be most suitable for completing the task, along with reasons for each choice. The response should only include the JSON object, without any additional explanatory text.

Please provide the recommended tools in the following format:
{{"tools": [{{"tool_id": tool_id, "reason": reason_for_recommending }}]}}

Task: {user_input}

Tools List (in JSON format):
{tools_string}
"""


def select_packs(task_description: str, llm: BaseChatModel) -> list[str]:
    """Given a user input describing the task they wish to accomplish, return a list of Pack IDs that the given LLM
    thinks will be suitable for this task.

    This is good for a "pre-processing" step after receiving the task but before trying to solve it. This allows you
    to benefit from the wide tool selection while keeping your token usage low.

    You can then further filter, install the packs if desired, and then fetch them using get_pack().

    Args:
        task_description (str): A description of the task to be used when selecting tools
        llm (BaseChatModel): An LLM which will be used to evaluate the selection

    Returns:
        list[str]: A list of selected Pack IDs
    """

    packs = pack_search("")
    pack_summaries = []
    for i, pack in enumerate(packs):
        # Use a number instead of a tool ID because the AI can mismatch tool names and repos
        pseudo_id = i + 1
        pack_summaries.append(
            {"tool_id": pseudo_id, "name": pack.name, "description": pack.description, "arguments": pack.run_args}
        )

    prompt = TOOL_SELECTION_PROMPT_TEMPLATE.format(user_input=task_description, tools_string=json.dumps(pack_summaries))

    response = ask_llm(prompt, llm)

    # TODO Handle json errors, perhaps a retry
    try:
        tools = json.loads(response.content)
    except json.JSONDecodeError as e:
        return []

    selected_packs = []
    for selected_tool in tools.get("tools"):
        tool_index = int(selected_tool.get("tool_id"))
        selected_packs.append(packs[tool_index - 1].pack_id)

    return selected_packs


def ask_llm(prompt: str, llm: BaseChatModel) -> BaseMessage:
    """Encapsulate the OpenAI specific stuff to easier support other frameworks in the future"""
    message = SystemMessage(content=prompt)

    response = llm(messages=[message])

    return response
